@@ -40,7 +40,9 @@
 #include "utils/rel.h"
 #include "utils/snapmgr.h"
 #include "utils/syscache.h"
+#if PG_VERSION_NUM < 120000
 #include "utils/tqual.h"
+#endif
 #include "utils/typcache.h"
 
 #include "pglogical_conflict.h"
@@ -135,7 +137,9 @@ static bool
 find_index_tuple(ScanKey skey, Relation rel, Relation idxrel,
 				 LockTupleMode lockmode, TupleTableSlot *slot)
 {
+#if PG_VERSION_NUM < 120000
 	HeapTuple	scantuple;
+#endif
 	bool		found;
 	IndexScanDesc scan;
 	SnapshotData snap;
@@ -157,11 +161,16 @@ retry:
 	index_rescan(scan, skey, IndexRelationGetNumberOfKeyAttributes(idxrel),
 				 NULL, 0);
 
+#if PG_VERSION_NUM >= 120000
+	if (index_getnext_slot(scan, ForwardScanDirection, slot))
+#else
 	if ((scantuple = index_getnext(scan, ForwardScanDirection)) != NULL)
+#endif
 	{
 		found = true;
-		/* FIXME: Improve TupleSlot to not require copying the whole tuple */
+#if PG_VERSION_NUM < 120000
 		ExecStoreTuple(scantuple, slot, InvalidBuffer, false);
+#endif
 		ExecMaterializeSlot(slot);
 
 		/*
@@ -182,15 +191,29 @@ retry:
 	/* Matching tuple found, no concurrent txns modifying it */
 	if (found)
 	{
+#if PG_VERSION_NUM >= 120000
+		TM_FailureData tmfd;
+		TM_Result res;
+#else
 		Buffer buf;
 		HeapUpdateFailureData hufd;
 		HTSU_Result res;
 		HeapTupleData locktup;
 
 		ItemPointerCopy(&slot->tts_tuple->t_self, &locktup.t_self);
+#endif
 
 		PushActiveSnapshot(GetLatestSnapshot());
 
+#if PG_VERSION_NUM >= 120000
+		res = table_tuple_lock(rel, &(slot->tts_tid), GetLatestSnapshot(),
+							   slot,
+							   GetCurrentCommandId(false),
+							   lockmode,
+							   LockWaitBlock,
+							   0 /* don't follow updates */ ,
+							   &tmfd);
+#else
 		res = heap_lock_tuple(rel, &locktup, GetCurrentCommandId(false),
 							  lockmode,
 							  false /* wait */,
@@ -198,15 +221,24 @@ retry:
 							  &buf, &hufd);
 		/* the tuple slot already has the buffer pinned */
 		ReleaseBuffer(buf);
+#endif
 
 		PopActiveSnapshot();
 
 		switch (res)
 		{
+#if PG_VERSION_NUM >= 120000
+			case TM_Ok:
+#else
 			case HeapTupleMayBeUpdated:
+#endif
 				/* lock was successfully acquired */
 				break;
+#if PG_VERSION_NUM >= 120000
+			case TM_Updated:
+#else
 			case HeapTupleUpdated:
+#endif
 				/*
 				 * We lost a race between when we looked up the tuple and
 				 * checked for concurrent modifying txns and when we tried to
@@ -377,7 +409,11 @@ pglogical_tuple_find_conflict(EState *estate, PGLogicalTupleData *tuple,
 
 		if (found)
 		{
+#if PG_VERSION_NUM >= 120000
+			ItemPointerCopy(&outslot->tts_tid, &conflicting_tid);
+#else
 			ItemPointerCopy(&outslot->tts_tuple->t_self, &conflicting_tid);
+#endif
 			conflict_idx = RelationGetRelid(idxrel);
 			break;
 		}
@@ -724,17 +760,10 @@ static void
 tuple_to_stringinfo(StringInfo s, TupleDesc tupdesc, HeapTuple tuple)
 {
 	int			natt;
-	Oid			oid;
 	bool		first = true;
 
 	static const int MAX_CONFLICT_LOG_ATTR_LEN = 40;
 
-	/* print oid of tuple, it's not included in the TupleDesc */
-	if ((oid = HeapTupleHeaderGetOid(tuple->t_data)) != InvalidOid)
-	{
-		appendStringInfo(s, "oid[oid]:%u", oid);
-	}
-
 	/* print all columns individually */
 	for (natt = 0; natt < tupdesc->natts; natt++)
 	{
