@@ -21,6 +21,9 @@
 
 #include "postgres.h"
 
+#if PG_VERSION_NUM >= 120000
+#include "access/heapam.h"
+#endif
 #include "access/htup_details.h"
 #include "access/xact.h"
 #include "catalog/dependency.h"
@@ -82,7 +85,9 @@
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
 #include "utils/syscache.h"
+#if PG_VERSION_NUM < 120000
 #include "utils/tqual.h"
+#endif
 
 #include "pglogical_dependency.h"
 #include "pglogical_sync.h"
@@ -305,7 +310,7 @@ pglogical_recordMultipleDependencies(const ObjectAddress *depender,
 	if (nreferenced <= 0)
 		return;					/* nothing to do */
 
-	dependDesc = heap_open(get_pglogical_depend_rel_oid(),
+	dependDesc = table_open(get_pglogical_depend_rel_oid(),
 						   RowExclusiveLock);
 
 	memset(nulls, false, sizeof(nulls));
@@ -333,7 +338,7 @@ pglogical_recordMultipleDependencies(const ObjectAddress *depender,
 		heap_freetuple(tup);
 	}
 
-	heap_close(dependDesc, RowExclusiveLock);
+	table_close(dependDesc, RowExclusiveLock);
 }
 
 
@@ -461,7 +466,7 @@ findDependentObjects(const ObjectAddress *object,
 #endif
 				/* no problem */
 				break;
-#if PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000 && PG_VERSION_NUM < 120000
 			case DEPENDENCY_INTERNAL_AUTO:
 #endif
 			case DEPENDENCY_INTERNAL:
@@ -545,7 +550,7 @@ findDependentObjects(const ObjectAddress *object,
 				 * other words, we don't follow the links back to the owning
 				 * object.
 				 */
-#if PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000 && PG_VERSION_NUM < 120000
 				if (foundDep->deptype == DEPENDENCY_INTERNAL_AUTO)
 					break;
 #endif
@@ -683,7 +688,7 @@ findDependentObjects(const ObjectAddress *object,
 				break;
 #endif
 			case DEPENDENCY_INTERNAL:
-#if PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000 && PG_VERSION_NUM < 120000
 			case DEPENDENCY_INTERNAL_AUTO:
 #endif
 				subflags = DEPFLAG_INTERNAL;
@@ -1913,7 +1918,7 @@ void pglogical_tryDropDependencies(const ObjectAddress *object,
 	 * We save some cycles by opening pglogical_depend just once and passing the
 	 * Relation pointer down to all the recursive deletion steps.
 	 */
-	depRel = heap_open(get_pglogical_depend_rel_oid(), RowExclusiveLock);
+	depRel = table_open(get_pglogical_depend_rel_oid(), RowExclusiveLock);
 
 	/*
 	 * Construct a list of objects to delete (ie, the given object plus
@@ -1953,7 +1958,7 @@ void pglogical_tryDropDependencies(const ObjectAddress *object,
 	/* And clean up */
 	free_object_addresses(targetObjects);
 
-	heap_close(depRel, RowExclusiveLock);
+	table_close(depRel, RowExclusiveLock);
 }
 
 /*
