@@ -50,6 +50,7 @@
 #include "replication/reorderbuffer.h"
 #include "replication/slot.h"
 
+#include "storage/ipc.h"
 #include "storage/latch.h"
 #include "storage/proc.h"
 
@@ -132,6 +133,7 @@ PG_FUNCTION_INFO_V1(pglogical_dependency_check_trigger);
 PG_FUNCTION_INFO_V1(pglogical_gen_slot_name);
 PG_FUNCTION_INFO_V1(pglogical_node_info);
 PG_FUNCTION_INFO_V1(pglogical_show_repset_table_info);
+PG_FUNCTION_INFO_V1(pglogical_show_repset_table_info_by_target);
 PG_FUNCTION_INFO_V1(pglogical_table_data_filtered);
 
 /* Information */
@@ -388,10 +390,11 @@ pglogical_create_subscription(PG_FUNCTION_ARGS)
 	char				   *sub_name = NameStr(*PG_GETARG_NAME(0));
 	char				   *provider_dsn = text_to_cstring(PG_GETARG_TEXT_PP(1));
 	ArrayType			   *rep_set_names = PG_GETARG_ARRAYTYPE_P(2);
-	bool					sync_structure = PG_GETARG_BOOL(3);
+	char					*sync_structure = text_to_cstring(PG_GETARG_TEXT_PP(3));
 	bool					sync_data = PG_GETARG_BOOL(4);
 	ArrayType			   *forward_origin_names = PG_GETARG_ARRAYTYPE_P(5);
 	Interval			   *apply_delay = PG_GETARG_INTERVAL_P(6);
+	bool					force_text_transfer = PG_GETARG_BOOL(7);
 	PGconn				   *conn;
 	PGLogicalSubscription	sub;
 	PGLogicalSyncStatus		sync;
@@ -506,16 +509,21 @@ pglogical_create_subscription(PG_FUNCTION_ARGS)
 				  origin.name, sub_name);
 	sub.slot_name = pstrdup(NameStr(slot_name));
 	sub.apply_delay = apply_delay;
+	sub.force_text_transfer = force_text_transfer;
 
 	create_subscription(&sub);
 
 	/* Create synchronization status for the subscription. */
 	memset(&sync, 0, sizeof(PGLogicalSyncStatus));
 
-	if (sync_structure && sync_data)
+	if (SyncStructureAll(sync_structure) && sync_data)
 		sync.kind = SYNC_KIND_FULL;
-	else if (sync_structure)
+	else if (SyncStructureRelOnly(sync_structure) && sync_data)
+		sync.kind = SYNC_KIND_FULL_REL;
+	else if (SyncStructureAll(sync_structure))
 		sync.kind = SYNC_KIND_STRUCTURE;
+	else if (SyncStructureRelOnly(sync_structure))
+		sync.kind = SYNC_KIND_STRUCTURE_REL;
 	else if (sync_data)
 		sync.kind = SYNC_KIND_DATA;
 	else
@@ -585,6 +593,8 @@ pglogical_drop_subscription(PG_FUNCTION_ARGS)
 		/* Wait for the apply to die. */
 		for (;;)
 		{
+			int rc;
+
 			LWLockAcquire(PGLogicalCtx->lock, LW_EXCLUSIVE);
 			apply = pglogical_apply_find(MyDatabaseId, sub->id);
 			if (!pglogical_worker_running(apply))
@@ -596,8 +606,11 @@ pglogical_drop_subscription(PG_FUNCTION_ARGS)
 
 			CHECK_FOR_INTERRUPTS();
 
-			(void) WaitLatch(&MyProc->procLatch,
-							 WL_LATCH_SET | WL_TIMEOUT, 1000L);
+			rc = WaitLatch(&MyProc->procLatch,
+						   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH, 1000L);
+
+			if (rc & WL_POSTMASTER_DEATH)
+				proc_exit(1);
 
 			ResetLatch(&MyProc->procLatch);
 		}
@@ -821,8 +834,8 @@ pglogical_alter_subscription_synchronize(PG_FUNCTION_ARGS)
 			/* We might delete the cell so advance it now. */
 			next = lnext(llc);
 
-			if (namestrcmp(&tablesync->nspname, remoterel->nspname) == 0 &&
-				namestrcmp(&tablesync->relname, remoterel->relname) == 0)
+			if (namestrcmp(&tablesync->nspname, remoterel->nsptarget) == 0 &&
+				namestrcmp(&tablesync->relname, remoterel->reltarget) == 0)
 			{
 				oldsync = tablesync;
 				local_tables = list_delete_cell(local_tables, llc, prev);
@@ -839,13 +852,13 @@ pglogical_alter_subscription_synchronize(PG_FUNCTION_ARGS)
 			memset(&newsync, 0, sizeof(PGLogicalSyncStatus));
 			newsync.kind = SYNC_KIND_DATA;
 			newsync.subid = sub->id;
-			namestrcpy(&newsync.nspname, remoterel->nspname);
-			namestrcpy(&newsync.relname, remoterel->relname);
+			namestrcpy(&newsync.nspname, remoterel->nsptarget);
+			namestrcpy(&newsync.relname, remoterel->reltarget);
 			newsync.status = SYNC_STATUS_INIT;
 			create_local_sync_status(&newsync);
 
 			if (truncate)
-				truncate_table(remoterel->nspname, remoterel->relname);
+				truncate_table(remoterel->nsptarget, remoterel->reltarget);
 		}
 	}
 
@@ -883,7 +896,7 @@ pglogical_alter_subscription_resynchronize_table(PG_FUNCTION_ARGS)
 	char				   *nspname,
 						   *relname;
 
-	rel = heap_open(reloid, AccessShareLock);
+	rel = table_open(reloid, AccessShareLock);
 
 	nspname = get_namespace_name(RelationGetNamespace(rel));
 	relname = RelationGetRelationName(rel);
@@ -893,6 +906,7 @@ pglogical_alter_subscription_resynchronize_table(PG_FUNCTION_ARGS)
 	if (oldsync)
 	{
 		if (oldsync->status != SYNC_STATUS_READY &&
+			oldsync->status != SYNC_STATUS_SYNCDONE &&
 			oldsync->status != SYNC_STATUS_NONE)
 			elog(ERROR, "table %s.%s is already being synchronized",
 				 nspname, relname);
@@ -913,7 +927,7 @@ pglogical_alter_subscription_resynchronize_table(PG_FUNCTION_ARGS)
 		create_local_sync_status(&newsync);
 	}
 
-	heap_close(rel, NoLock);
+	table_close(rel, NoLock);
 
 	if (truncate)
 		truncate_table(nspname, relname);
@@ -1319,6 +1333,9 @@ parse_row_filter(Relation rel, char *row_filter_str)
 	pstate = make_parsestate(NULL);
 	rte = addRangeTableEntryForRelation(pstate,
 										rel,
+#if PG_VERSION_NUM >= 120000
+										AccessShareLock,
+#endif
 										NULL,
 										false,
 										true);
@@ -1367,6 +1384,8 @@ pglogical_replication_set_add_table(PG_FUNCTION_ARGS)
 	PGLogicalLocalNode *node;
 	char			   *nspname;
 	char			   *relname;
+	char			   *nsptarget;
+	char			   *reltarget;
 	StringInfoData		json;
 
 	/* Proccess for required parameters. */
@@ -1398,7 +1417,7 @@ pglogical_replication_set_add_table(PG_FUNCTION_ARGS)
 	 * Make sure the relation exists (lock mode has to be the same one as
 	 * in replication_set_add_relation).
 	 */
-	rel = heap_open(reloid, ShareRowExclusiveLock);
+	rel = table_open(reloid, ShareRowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	nspname = get_namespace_name(RelationGetNamespace(rel));
@@ -1444,25 +1463,34 @@ pglogical_replication_set_add_table(PG_FUNCTION_ARGS)
 									  text_to_cstring(PG_GETARG_TEXT_PP(4)));
 	}
 
-	replication_set_add_table(repset->id, reloid, att_list, row_filter);
+	if (!PG_ARGISNULL(5))
+	  nsptarget = NameStr(*PG_GETARG_NAME(5));
+	else
+	  nsptarget = pstrdup(nspname);
+	if (!PG_ARGISNULL(6))
+	  reltarget = NameStr(*PG_GETARG_NAME(6));
+	else
+	  reltarget = pstrdup(relname);
+
+	replication_set_add_table(repset->id, reloid, att_list, row_filter,
+							  nsptarget, reltarget);
 
 	if (synchronize)
 	{
 		/* It's easier to construct json manually than via Jsonb API... */
 		initStringInfo(&json);
 		appendStringInfo(&json, "{\"schema_name\": ");
-		escape_json(&json, nspname);
+		escape_json(&json, nsptarget);
 		appendStringInfo(&json, ",\"table_name\": ");
-		escape_json(&json, relname);
+		escape_json(&json, reltarget);
 		appendStringInfo(&json, "}");
-
-		/* Queue the truncate for replication. */
+		/* Queue the synchronize request for replication. */
 		queue_message(list_make1(repset->name), GetUserId(),
 					  QUEUE_COMMAND_TYPE_TABLESYNC, json.data);
 	}
 
 	/* Cleanup. */
-	heap_close(rel, NoLock);
+	table_close(rel, ShareRowExclusiveLock);
 
 	PG_RETURN_BOOL(true);
 }
@@ -1481,6 +1509,8 @@ pglogical_replication_set_add_sequence(PG_FUNCTION_ARGS)
 	PGLogicalLocalNode *node;
 	char			   *nspname;
 	char			   *relname;
+	char			   *nsptarget;
+	char			   *reltarget;
 	StringInfoData		json;
 
 	node = check_local_node(true);
@@ -1493,21 +1523,29 @@ pglogical_replication_set_add_sequence(PG_FUNCTION_ARGS)
 	 * Make sure the relation exists (lock mode has to be the same one as
 	 * in replication_set_add_relation).
 	 */
-	rel = heap_open(reloid, ShareRowExclusiveLock);
+	rel = table_open(reloid, ShareRowExclusiveLock);
 
-	replication_set_add_seq(repset->id, reloid);
+	nspname = get_namespace_name(RelationGetNamespace(rel));
+	relname = RelationGetRelationName(rel);
+	if (!PG_ARGISNULL(3))
+	    nsptarget = NameStr(*PG_GETARG_NAME(3));
+	else
+	  nsptarget = pstrdup(nspname);
+	if (!PG_ARGISNULL(4))
+		reltarget = NameStr(*PG_GETARG_NAME(4));
+	else
+	  reltarget = pstrdup(relname);
+
+	replication_set_add_seq(repset->id, reloid, nsptarget, reltarget);
 
 	if (synchronize)
 	{
-		nspname = get_namespace_name(RelationGetNamespace(rel));
-		relname = RelationGetRelationName(rel);
-
 		/* It's easier to construct json manually than via Jsonb API... */
 		initStringInfo(&json);
 		appendStringInfo(&json, "{\"schema_name\": ");
-		escape_json(&json, nspname);
+		escape_json(&json, nsptarget);
 		appendStringInfo(&json, ",\"sequence_name\": ");
-		escape_json(&json, relname);
+		escape_json(&json, reltarget);
         appendStringInfo(&json, ",\"last_value\": \""INT64_FORMAT"\"",
 								 sequence_get_last_value(reloid));
 		appendStringInfo(&json, "}");
@@ -1518,7 +1556,7 @@ pglogical_replication_set_add_sequence(PG_FUNCTION_ARGS)
 	}
 
 	/* Cleanup. */
-	heap_close(rel, NoLock);
+	table_close(rel, NoLock);
 
 	PG_RETURN_BOOL(true);}
 
@@ -1547,7 +1585,7 @@ pglogical_replication_set_add_all_relations(Name repset_name,
 	existing_relations = list_concat_unique_oid(existing_relations,
 												replication_set_get_seqs(repset->id));
 
-	rel = heap_open(RelationRelationId, RowExclusiveLock);
+	rel = table_open(RelationRelationId, RowExclusiveLock);
 
 	foreach (lc, textarray_to_list(nsp_names))
 	{
@@ -1582,9 +1620,11 @@ pglogical_replication_set_add_all_relations(Name repset_name,
 			if (!list_member_oid(existing_relations, reloid))
 			{
 				if (relkind == RELKIND_RELATION)
-					replication_set_add_table(repset->id, reloid, NIL, NULL);
+					replication_set_add_table(repset->id, reloid, NIL, NULL,
+											  NULL, NULL);
 				else
-					replication_set_add_seq(repset->id, reloid);
+					replication_set_add_seq(repset->id, reloid,
+											NULL, NULL);
 
 				if (synchronize)
 				{
@@ -1627,7 +1667,7 @@ pglogical_replication_set_add_all_relations(Name repset_name,
 		systable_endscan(sysscan);
 	}
 
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 	PG_RETURN_BOOL(true);
 }
@@ -1648,7 +1688,7 @@ pglogical_replication_set_add_all_tables(PG_FUNCTION_ARGS)
 }
 
 /*
- * Add replication set / table mapping based on schemas.
+ * Add replication set / sequence mapping based on schemas.
  */
 Datum
 pglogical_replication_set_add_all_sequences(PG_FUNCTION_ARGS)
@@ -1893,20 +1933,26 @@ pglogical_node_info(PG_FUNCTION_ARGS)
  * This is called by downstream sync worker on the upstream to obtain
  * info needed to do initial synchronization correctly. Be careful
  * about changing it, as it must be upward- and downward-compatible.
+ *
+ * it's not used starting with pglogical 2.3 (and can be remove if we drop
+ * support to pre-2.3 in the future.
+ *
+ * It will return only tables with same name and schema on subscriber and
+ * provider
  */
 Datum
 pglogical_show_repset_table_info(PG_FUNCTION_ARGS)
 {
 	Oid			reloid = PG_GETARG_OID(0);
- 	ArrayType  *rep_set_names = PG_GETARG_ARRAYTYPE_P(1);
+	ArrayType  *rep_set_names =	PG_GETARG_ARRAYTYPE_P(1);
 	Relation	rel;
 	List	   *replication_sets;
 	TupleDesc	reldesc;
 	TupleDesc	rettupdesc;
 	int			i;
 	List	   *att_list = NIL;
-	Datum		values[5];
-	bool		nulls[5];
+	Datum		values[7];
+	bool		nulls[7];
 	char	   *nspname;
 	char	   *relname;
 	HeapTuple	htup;
@@ -1920,7 +1966,7 @@ pglogical_show_repset_table_info(PG_FUNCTION_ARGS)
 		elog(ERROR, "return type must be a row type");
 	rettupdesc = BlessTupleDesc(rettupdesc);
 
-	rel = heap_open(reloid, AccessShareLock);
+	rel = table_open(reloid, AccessShareLock);
 	reldesc = RelationGetDescr(rel);
 	replication_sets = textarray_to_list(rep_set_names);
 	replication_sets = get_replication_sets(node->node->id,
@@ -1932,7 +1978,7 @@ pglogical_show_repset_table_info(PG_FUNCTION_ARGS)
 
 	/* Build the replication info for the table. */
 	tableinfo = get_table_replication_info(node->node->id, rel,
-										   replication_sets);
+												  replication_sets);
 
 	/* Build the column list. */
 	for (i = 0; i < reldesc->natts; i++)
@@ -1959,14 +2005,133 @@ pglogical_show_repset_table_info(PG_FUNCTION_ARGS)
 	values[2] = CStringGetTextDatum(relname);
 	values[3] = PointerGetDatum(strlist_to_textarray(att_list));
 	values[4] = BoolGetDatum(list_length(tableinfo->row_filter) > 0);
+	values[5] = CStringGetTextDatum(tableinfo->nsptarget);
+	values[6] = CStringGetTextDatum(tableinfo->reltarget);
 
 	htup = heap_form_tuple(rettupdesc, values, nulls);
 
-	heap_close(rel, NoLock);
+	table_close(rel, NoLock);
 
 	PG_RETURN_DATUM(HeapTupleGetDatum(htup));
 }
 
+/*
+ * Get replication info about table, by target name
+ *
+ * This is called by downstream sync worker on the upstream to obtain
+ * info needed to do initial synchronization correctly. Be careful
+ * about changing it, as it must be upward- and downward-compatible.
+ */
+Datum
+pglogical_show_repset_table_info_by_target(PG_FUNCTION_ARGS)
+{
+	ArrayType  *rep_set_names = PG_GETARG_ARRAYTYPE_P(2);
+	RangeVar   *target;
+	char	   *nsptarget;
+	char	   *reltarget;
+	Relation	rel;
+	List	   *replication_sets;
+	TupleDesc	reldesc;
+	int			i;
+	Datum		values[7];
+	bool		nulls[7];
+	char	   *nspname;
+	char	   *relname;
+	PGLogicalLocalNode *node;
+	List *tablesinfo = NIL;
+	ListCell		   *lc;
+	ReturnSetInfo	   *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
+	TupleDesc			tupdesc;
+	Tuplestorestate	   *tupstore;
+	MemoryContext		per_query_ctx;
+	MemoryContext		oldcontext;
+
+	if (PG_ARGISNULL(0))
+			elog(ERROR,"Schema target name required");
+	nsptarget = NameStr(*PG_GETARG_NAME(0));
+	if (PG_ARGISNULL(1))
+			elog(ERROR,"Table target name required");
+	reltarget = NameStr(*PG_GETARG_NAME(1));
+
+	node = get_local_node(false, false);
+
+	target = makeRangeVar(nsptarget, reltarget, -1);
+
+	replication_sets = textarray_to_list(rep_set_names);
+	replication_sets = get_replication_sets(node->node->id,
+											replication_sets,
+											false);
+
+	/* Build the replication info for the table. */
+	tablesinfo = get_table_replication_info_by_target(node->node->id,
+													  target->schemaname,
+													  target->relname,
+													  replication_sets);
+
+	/* Switch into long-lived context to construct returned data structures */
+	per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
+	oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
+	/* Build a tuple descriptor for our result type */
+	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+		elog(ERROR, "return type must be a row type");
+	tupdesc = BlessTupleDesc(tupdesc);
+
+	tupstore = tuplestore_begin_heap(true, false, work_mem);
+	rsinfo->returnMode = SFRM_Materialize;
+	rsinfo->setResult = tupstore;
+	rsinfo->setDesc = tupdesc;
+
+	MemoryContextSwitchTo(oldcontext);
+
+	foreach (lc, tablesinfo)
+	{
+			List	   *att_list = NIL;
+			PGLogicalTableRepInfo *tableinfo = (PGLogicalTableRepInfo *) lfirst(lc);
+
+			rel = table_open(tableinfo->reloid, AccessShareLock);
+			reldesc = RelationGetDescr(rel);
+
+			nspname = get_namespace_name(RelationGetNamespace(rel));
+			relname = RelationGetRelationName(rel);
+
+			/* Build the column list. */
+			for (i = 0; i < reldesc->natts; i++)
+			{
+					Form_pg_attribute att = TupleDescAttr(reldesc,i);
+
+					/* Skip dropped columns. */
+					if (att->attisdropped)
+							continue;
+
+					/* Skip filtered columns if any. */
+					if (tableinfo->att_list &&
+						!bms_is_member(att->attnum - FirstLowInvalidHeapAttributeNumber,
+									   tableinfo->att_list))
+							continue;
+
+					att_list = lappend(att_list, NameStr(att->attname));
+			}
+
+			/* And now build the result. */
+			memset(nulls, false, sizeof(nulls));
+			values[0] = ObjectIdGetDatum(RelationGetRelid(rel));
+			values[1] = CStringGetTextDatum(nspname);
+			values[2] = CStringGetTextDatum(relname);
+			values[3] = PointerGetDatum(strlist_to_textarray(att_list));
+			values[4] = BoolGetDatum(list_length(tableinfo->row_filter) > 0);
+			values[5] = CStringGetTextDatum(tableinfo->nsptarget);
+			values[6] = CStringGetTextDatum(tableinfo->reltarget);
+
+			tuplestore_putvalues(tupstore, tupdesc, values, nulls);
+			table_close(rel, NoLock);
+	}
+
+	// XXX not required
+	tuplestore_donestoring(tupstore);
+
+	PG_RETURN_VOID();
+}
 
 /*
  * Decide if to return tuple or not.
@@ -1976,7 +2141,7 @@ filter_tuple(HeapTuple htup, ExprContext *econtext, List *row_filter_list)
 {
 	ListCell	   *lc;
 
-	ExecStoreTuple(htup, econtext->ecxt_scantuple, InvalidBuffer, false);
+	ExecStoreHeapTuple(htup, econtext->ecxt_scantuple, false);
 
 	foreach (lc, row_filter_list)
 	{
@@ -2016,7 +2181,7 @@ pglogical_table_data_filtered(PG_FUNCTION_ARGS)
 	ListCell   *lc;
 	TupleDesc	tupdesc;
 	TupleDesc	reltupdesc;
-	HeapScanDesc scandesc;
+	TableScanDesc scandesc;
 	HeapTuple	htup;
 	List	   *row_filter_list = NIL;
 	EState		   *estate;
@@ -2083,7 +2248,7 @@ pglogical_table_data_filtered(PG_FUNCTION_ARGS)
 	MemoryContextSwitchTo(oldcontext);
 
 	/* Check output type and table row type are the same. */
-	rel = heap_open(reloid, AccessShareLock);
+	rel = table_open(reloid, AccessShareLock);
 	reltupdesc = RelationGetDescr(rel);
 	if (!equalTupleDescs(tupdesc, reltupdesc))
 		ereport(ERROR,
@@ -2114,7 +2279,7 @@ pglogical_table_data_filtered(PG_FUNCTION_ARGS)
 
 
 	/* Scan the table. */
-	scandesc = heap_beginscan(rel, GetActiveSnapshot(), 0, NULL);
+	scandesc = table_beginscan(rel, GetActiveSnapshot(), 0, NULL);
 
 	while (HeapTupleIsValid(htup = heap_getnext(scandesc, ForwardScanDirection)))
 	{
@@ -2129,7 +2294,7 @@ pglogical_table_data_filtered(PG_FUNCTION_ARGS)
 	FreeExecutorState(estate);
 
 	heap_endscan(scandesc);
-	heap_close(rel, NoLock);
+	table_close(rel, NoLock);
 
 	PG_RETURN_NULL();
 }
@@ -2162,6 +2327,7 @@ pglogical_wait_for_sync_complete(char *subscription_name, char *relnamespace, ch
 		PGLogicalSyncStatus	   *subsync;
 		List				   *tables;
 		bool					isdone = false;
+		int						rc;
 
 		/* We need to see the latest rows */
 		PushActiveSnapshot(GetLatestSnapshot());
@@ -2173,19 +2339,19 @@ pglogical_wait_for_sync_complete(char *subscription_name, char *relnamespace, ch
 		if (isdone)
 		{
 			/*
-			 * Subscription its self is synced, but what about separately
+			 * Subscription itself is synced, but what about separately
 			 * synced tables?
 			 */
 			if (relname != NULL)
 			{
-				PGLogicalSyncStatus *table = get_table_sync_status(sub->id, relnamespace, relname, true);
+				PGLogicalSyncStatus *table = get_table_sync_status(sub->id, relnamespace, relname, false);
 				isdone = table && table->status == SYNC_STATUS_READY;
 				free_sync_status(table);
 			}
 			else
 			{
 				/*
-				 * This is plenty inefficient and we should probably just do a direct catalog
+				 * XXX This is plenty inefficient and we should probably just do a direct catalog
 				 * scan, but meh, it hardly has to be fast.
 				 */
 				ListCell *lc;
@@ -2208,8 +2374,11 @@ pglogical_wait_for_sync_complete(char *subscription_name, char *relnamespace, ch
 		CHECK_FOR_INTERRUPTS();
 
 		/* some kind of backoff could be useful here */
-		(void) WaitLatch(&MyProc->procLatch,
-						 WL_LATCH_SET | WL_TIMEOUT, 200L);
+		rc = WaitLatch(&MyProc->procLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH, 200L);
+
+		if (rc & WL_POSTMASTER_DEATH)
+			proc_exit(1);
 
 		ResetLatch(&MyProc->procLatch);
 	} while (1);
