@@ -87,12 +87,13 @@ static PGLogicalSyncWorker	   *MySyncWorker = NULL;
 
 static void
 dump_structure(PGLogicalSubscription *sub, const char *destfile,
-			   const char *snapshot)
+	       const char *snapshot, const char kind)
 {
 	char		pg_dump[MAXPGPATH];
 	uint32		version;
 	int			res;
 	StringInfoData	schema_filter;
+	StringInfoData	table_filter;
 	StringInfoData	command;
 
 	if (find_other_exec_version(my_exec_path, PGDUMP_BINARY, &version, pg_dump) != 0)
@@ -112,10 +113,45 @@ dump_structure(PGLogicalSubscription *sub, const char *destfile,
 		appendStringInfoString(&schema_filter, " -N pglogical_origin");
 	CommitTransactionCommand();
 
+	/*
+	 * Filter tables and sequences using "pg_dump -t" switch, based on
+	 * replication set tables and sequences from origin node.
+	 *
+	 * Creates new connection to origin.
+	 *
+	 * XXX CREATE SCHEMA is not done by pg_dump when using -t switch
+	 *
+	 * XXX add -T switch option ?
+	 *
+	 */
+	initStringInfo(&table_filter);
+	if (SyncKindStructureRelations(kind))
+	{
+		List *objects;
+		ListCell   *lc;
+
+		objects = list_replication_sets_objects(sub->origin_if->dsn,
+							sub->name, snapshot,
+							sub->replication_sets);
+
+		/* Get every object and add it to the -t switch */
+		foreach (lc, objects)
+		{
+			char	   *object_name = lfirst(lc);
+			appendStringInfo(&table_filter, " -t %s",
+					 object_name);
+		}
+	}
+
 	initStringInfo(&command);
-	appendStringInfo(&command, "\"%s\" --snapshot=\"%s\" %s -s -F c -f \"%s\" \"%s\"",
-					 pg_dump, snapshot, schema_filter.data, destfile,
-					 sub->origin_if->dsn);
+#if PG_VERSION_NUM >= 90600
+	appendStringInfo(&command, "\"%s\" --strict-names --snapshot=\"%s\" %s %s -s -F c -f \"%s\" \"%s\"",
+#else
+	appendStringInfo(&command, "\"%s\" --snapshot=\"%s\" %s %s -s -F c -f \"%s\" \"%s\"",
+#endif
+					 pg_dump, snapshot,
+					 schema_filter.data, table_filter.data,
+					 destfile, sub->origin_if->dsn);
 
 	res = system(command.data);
 	if (res != 0)
@@ -504,10 +540,10 @@ copy_table_data(PGconn *origin_conn, PGconn *target_conn,
 	/* Build COPY FROM query. */
 	resetStringInfo(&query);
 	appendStringInfo(&query, "COPY %s.%s ",
-					 PQescapeIdentifier(origin_conn, remoterel->nspname,
-										strlen(remoterel->nspname)),
-					 PQescapeIdentifier(origin_conn, remoterel->relname,
-										strlen(remoterel->relname)));
+					 PQescapeIdentifier(target_conn, remoterel->nsptarget,
+										strlen(remoterel->nsptarget)),
+					 PQescapeIdentifier(target_conn, remoterel->reltarget,
+										strlen(remoterel->reltarget)));
 	if (list_length(attnamelist))
 		appendStringInfo(&query, "(%s) ", attlist.data);
 	appendStringInfoString(&query, "FROM stdin");
@@ -554,6 +590,77 @@ copy_table_data(PGconn *origin_conn, PGconn *target_conn,
 	}
 
 	PQclear(res);
+
+	elog(INFO, "finished synchronization of data for table %s.%s",
+		 remoterel->nsptarget, remoterel->reltarget);
+}
+
+/*
+ * Returns the list of schema qualified name of replicated tables and sequences
+ * from provider.
+ *
+ *  Creates new connection to origin.
+ */
+List *
+list_replication_sets_objects(const char *dsn, const char *name, const char *snapshot, List *replication_sets)
+{
+	PGconn	   *origin_conn;
+	List	   *objects;
+	List	   *res = NIL;
+	ListCell   *lc;
+
+	/* Connect to origin node. */
+	origin_conn = pglogical_connect(dsn, name, "copy");
+	start_copy_origin_tx(origin_conn, snapshot);
+
+	/* Get tables from our replication sets from origin node. */
+	objects = pg_logical_get_remote_repset_tables(origin_conn,
+												 replication_sets);
+
+	/* And append them to the list */
+	foreach (lc, objects)
+	{
+		PGLogicalRemoteRel	*remoterel = lfirst(lc);
+		StringInfoData object;
+
+		initStringInfo(&object);
+		appendStringInfo(&object, "%s.%s",
+						 PQescapeLiteral(origin_conn, remoterel->nsptarget,
+											strlen(remoterel->nsptarget)),
+						 PQescapeLiteral(origin_conn, remoterel->reltarget,
+											strlen(remoterel->reltarget)));
+		res = lappend(res, object.data);
+
+		/* XXX probably not required here */
+		CHECK_FOR_INTERRUPTS();
+	}
+
+	/* Get sequences from our replication sets from origin node. */
+	objects = pg_logical_get_remote_repset_sequences(origin_conn,
+												 replication_sets);
+
+	/* And append them to the list */
+	foreach (lc, objects)
+	{
+		RangeVar	*rv = lfirst(lc);
+		StringInfoData object;
+
+		initStringInfo(&object);
+		appendStringInfo(&object, "%s.%s",
+				 PQescapeLiteral(origin_conn, rv->schemaname,
+											strlen(rv->schemaname)),
+						 PQescapeLiteral(origin_conn, rv->relname,
+											strlen(rv->relname)));
+		res = lappend(res, object.data);
+
+		/* XXX probably not required here */
+		CHECK_FOR_INTERRUPTS();
+	}
+
+	/* Finish the transaction and disconnect. */
+	finish_copy_origin_tx(origin_conn);
+
+	return res;
 }
 
 /*
@@ -569,7 +676,7 @@ copy_tables_data(char *sub_name, const char *origin_dsn,
 {
 	PGconn	   *origin_conn;
 	PGconn	   *target_conn;
-	ListCell   *lc;
+	ListCell   *lc, *lcr;
 
 	/* Connect to origin node. */
 	origin_conn = pglogical_connect(origin_dsn, sub_name, "copy");
@@ -583,12 +690,14 @@ copy_tables_data(char *sub_name, const char *origin_dsn,
 	foreach (lc, tables)
 	{
 		RangeVar	*rv = lfirst(lc);
-		PGLogicalRemoteRel	*remoterel;
-
-		remoterel = pg_logical_get_remote_repset_table(origin_conn, rv,
+        List		*remoterels = NIL;
+		remoterels = pg_logical_get_remote_repset_table(origin_conn, rv,
 													   replication_sets);
-
-		copy_table_data(origin_conn, target_conn, remoterel, replication_sets);
+        foreach(lcr, remoterels)
+        {
+          PGLogicalRemoteRel	*remoterel = lfirst(lcr);
+          copy_table_data(origin_conn, target_conn, remoterel, replication_sets);
+        }
 
 		CHECK_FOR_INTERRUPTS();
 	}
@@ -794,12 +903,12 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 					originid,
 					(uint32)(XactLastCommitEnd>>32), (uint32)XactLastCommitEnd);
 #if PG_VERSION_NUM >= 90500
-				replorigin_rel = heap_open(ReplicationOriginRelationId, RowExclusiveLock);
+				replorigin_rel = table_open(ReplicationOriginRelationId, RowExclusiveLock);
 #endif
 				replorigin_advance(originid, lsn, XactLastCommitEnd, true,
 								   true);
 #if PG_VERSION_NUM >= 90500
-				heap_close(replorigin_rel, RowExclusiveLock);
+				table_close(replorigin_rel, RowExclusiveLock);
 #endif
 
 				CommitTransactionCommand();
@@ -814,7 +923,7 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 					CommitTransactionCommand();
 
 					/* Dump structure to temp storage. */
-					dump_structure(sub, tmpfile, snapshot);
+					dump_structure(sub, tmpfile, snapshot, sync->kind);
 
 					/* Restore base pre-data structure (types, tables, etc). */
 					restore_structure(sub, tmpfile, "pre-data");
@@ -848,12 +957,13 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 						PGLogicalSyncStatus	   *oldsync;
 
 						oldsync = get_table_sync_status(sub->id,
-														remoterel->nspname,
-														remoterel->relname, true);
+														remoterel->nsptarget,
+														remoterel->reltarget,
+														true);
 						if (oldsync)
 						{
-							set_table_sync_status(sub->id, remoterel->nspname,
-												  remoterel->relname,
+							set_table_sync_status(sub->id, remoterel->nsptarget,
+												  remoterel->reltarget,
 												  SYNC_STATUS_READY,
 												  lsn);
 						}
@@ -863,8 +973,8 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 
 							newsync.kind = SYNC_KIND_FULL;
 							newsync.subid = sub->id;
-							namestrcpy(&newsync.nspname, remoterel->nspname);
-							namestrcpy(&newsync.relname, remoterel->relname);
+							namestrcpy(&newsync.nspname, remoterel->nsptarget);
+							namestrcpy(&newsync.relname, remoterel->reltarget);
 							newsync.status = SYNC_STATUS_READY;
 							newsync.statuslsn = lsn;
 							create_local_sync_status(&newsync);
@@ -975,12 +1085,12 @@ pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table,
 			(uint32)(XactLastCommitEnd>>32), (uint32)XactLastCommitEnd);
 
 #if PG_VERSION_NUM >= 90500
-		replorigin_rel = heap_open(ReplicationOriginRelationId, RowExclusiveLock);
+		replorigin_rel = table_open(ReplicationOriginRelationId, RowExclusiveLock);
 #endif
 		replorigin_advance(originid, *status_lsn, XactLastCommitEnd, true,
 						   true);
 #if PG_VERSION_NUM >= 90500
-		heap_close(replorigin_rel, RowExclusiveLock);
+		table_close(replorigin_rel, RowExclusiveLock);
 #endif
 
 		set_table_sync_status(sub->id, table->schemaname, table->relname,
@@ -1162,7 +1272,8 @@ pglogical_sync_main(Datum main_arg)
 	pglogical_identify_system(streamConn, NULL, NULL, NULL, NULL);
 
 	pglogical_start_replication(streamConn, MySubscription->slot_name,
-								status_lsn, "all", NULL, tablename);
+								status_lsn, "all", NULL, tablename,
+								MySubscription->force_text_transfer);
 
 	/* Leave it to standard apply code to do the replication. */
 	apply_work(streamConn);
@@ -1191,7 +1302,7 @@ create_local_sync_status(PGLogicalSyncStatus *sync)
 	bool		nulls[Natts_local_sync_state];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	/* Form a tuple. */
@@ -1220,7 +1331,7 @@ create_local_sync_status(PGLogicalSyncStatus *sync)
 
 	/* Cleanup. */
 	heap_freetuple(tup);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 }
 
 /* Remove subscription sync status record from catalog. */
@@ -1234,7 +1345,7 @@ drop_subscription_sync_status(Oid subid)
 	ScanKeyData		key[1];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 
 	ScanKeyInit(&key[0],
 				Anum_sync_subid,
@@ -1249,8 +1360,7 @@ drop_subscription_sync_status(Oid subid)
 
 	/* Cleanup. */
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
-
+	table_close(rel, RowExclusiveLock);
 }
 
 static PGLogicalSyncStatus *
@@ -1302,7 +1412,7 @@ get_subscription_sync_status(Oid subid, bool missing_ok)
 	TupleDesc		tupDesc;
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	ScanKeyInit(&key[0],
@@ -1323,7 +1433,7 @@ get_subscription_sync_status(Oid subid, bool missing_ok)
 		if (missing_ok)
 		{
 			systable_endscan(scan);
-			heap_close(rel, RowExclusiveLock);
+			table_close(rel, RowExclusiveLock);
 			return NULL;
 		}
 
@@ -1333,7 +1443,7 @@ get_subscription_sync_status(Oid subid, bool missing_ok)
 	sync = syncstatus_fromtuple(tuple, tupDesc);
 
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 	return sync;
 }
@@ -1354,7 +1464,7 @@ set_subscription_sync_status(Oid subid, char status)
 	bool			replaces[Natts_local_sync_state];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	ScanKeyInit(&key[0],
@@ -1389,7 +1499,7 @@ set_subscription_sync_status(Oid subid, char status)
 	/* Cleanup. */
 	heap_freetuple(newtup);
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 }
 
 /* Remove table sync status record from catalog. */
@@ -1403,7 +1513,7 @@ drop_table_sync_status(const char *nspname, const char *relname)
 	ScanKeyData		key[2];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 
 	ScanKeyInit(&key[0],
 				Anum_sync_nspname,
@@ -1422,8 +1532,7 @@ drop_table_sync_status(const char *nspname, const char *relname)
 
 	/* Cleanup. */
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
-
+	table_close(rel, RowExclusiveLock);
 }
 
 /* Remove table sync status record from catalog. */
@@ -1438,7 +1547,7 @@ drop_table_sync_status_for_sub(Oid subid, const char *nspname,
 	ScanKeyData		key[3];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 
 	ScanKeyInit(&key[0],
 				Anum_sync_subid,
@@ -1461,7 +1570,7 @@ drop_table_sync_status_for_sub(Oid subid, const char *nspname,
 
 	/* Cleanup. */
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 }
 
@@ -1482,7 +1591,7 @@ get_table_sync_status(Oid subid, const char *nspname, const char *relname,
 	ListCell	   *l;
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 
 	/* Find an index we can use to scan this catalog. */
 	indexes = RelationGetIndexList(rel);
@@ -1527,7 +1636,7 @@ get_table_sync_status(Oid subid, const char *nspname, const char *relname,
 		if (missing_ok)
 		{
 			systable_endscan(scan);
-			heap_close(rel, RowExclusiveLock);
+			table_close(rel, RowExclusiveLock);
 			return NULL;
 		}
 
@@ -1538,7 +1647,7 @@ get_table_sync_status(Oid subid, const char *nspname, const char *relname,
 	sync = syncstatus_fromtuple(tuple, tupDesc);
 
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 	return sync;
 }
@@ -1557,7 +1666,7 @@ get_unsynced_tables(Oid subid)
 	TupleDesc		tupDesc;
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	ScanKeyInit(&key[0],
@@ -1579,7 +1688,7 @@ get_unsynced_tables(Oid subid)
 	}
 
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 	return res;
 }
@@ -1598,7 +1707,7 @@ get_subscription_tables(Oid subid)
 	TupleDesc		tupDesc;
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	ScanKeyInit(&key[0],
@@ -1619,7 +1728,7 @@ get_subscription_tables(Oid subid)
 	}
 
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 
 	return res;
 }
@@ -1641,7 +1750,7 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
 	bool			replaces[Natts_local_sync_state];
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
-	rel = heap_openrv(rv, RowExclusiveLock);
+	rel = table_openrv(rv, RowExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	ScanKeyInit(&key[0],
@@ -1680,7 +1789,7 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
 	/* Cleanup. */
 	heap_freetuple(newtup);
 	systable_endscan(scan);
-	heap_close(rel, RowExclusiveLock);
+	table_close(rel, RowExclusiveLock);
 }
 
 /*
@@ -1693,7 +1802,7 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
  * in TopMemoryContext.
  */
 bool
-wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
+wait_for_sync_status_change(Oid subid, const char *nspname, const char *relname,
 							char desired_state, XLogRecPtr *lsn)
 {
 	int rc;
